import aiosqlite
import sqlite3
import time

from contextlib import asynccontextmanager
from pathlib import Path

from quart import request, jsonify, current_app as app
from quart.views import MethodView
from slugify import slugify

from csvapi.errors import APIError
from csvapi.utils import get_db_info

ROWS_LIMIT = 100
SQL_TIME_LIMIT_MS = 1000
DEFAULT_SHAPE = 'lists'


def prepare_connection(conn):
    # conn.row_factory = sqlite3.Row
    conn.text_factory = lambda x: str(x, 'utf-8', 'replace')


@asynccontextmanager
async def sqlite_timelimit(conn, ms):
    deadline = time.time() + (ms / 1000)
    # n is the number of SQLite virtual machine instructions that will be
    # executed between each check. It's hard to know what to pick here.
    # After some experimentation, I've decided to go with 1000 by default and
    # 1 for time limits that are less than 50ms
    n = 1000
    if ms < 50:
        n = 1

    def handler():
        if time.time() >= deadline:
            return 1

    await conn.set_progress_handler(handler, n)
    yield
    await conn.set_progress_handler(None, n)


class TableView(MethodView):

    async def execute(self, sql, db_info, params=None):
        """Executes sql against db_name in a thread"""
        dsn = 'file:{}?immutable=1'.format(db_info['db_path'])
        # specify uri=True to make sure `file:xxx` is supported,
        # however the backend sqlite is configured (eg default MacOS)
        async with aiosqlite.connect(dsn, uri=True) as conn:
            conn.text_factory = lambda x: str(x, 'utf-8', 'replace')
            # this will raise
            #  {"details": "interrupted",
            #  "error": "Error selecting data",}
            async with sqlite_timelimit(conn, SQL_TIME_LIMIT_MS):
                try:
                    async with conn.execute(sql, params or {}) as cursor:
                        rows = await cursor.fetchall()
                except Exception:
                    app.logger.error('ERROR: conn={}, sql = {}, params = {}'.format(
                        conn, repr(sql), params
                    ))
                    raise
            return rows, cursor.description

    def add_filters_to_sql(self, sql, filters):
        wheres = []
        params = {}
        for (f_key, f_value) in filters:
            comparator = f_key.split('__')[1]
            column = f_key.split('__')[0]
            normalized_column = slugify(column, separator='_')
            if comparator == 'exact':
                wheres.append(f"[{column}] = :filter_value_{normalized_column}")
                params[f'filter_value_{normalized_column}'] = f_value
            elif comparator == 'contains':
                wheres.append(f"[{column}] LIKE :filter_value_{normalized_column}")
                params[f'filter_value_{normalized_column}'] = f'%{f_value}%'
            else:
                app.logger.warning(f'Dropped unknown comparator in {f_key}')
        if wheres:
            sql += ' WHERE '
            sql += ' AND '.join(wheres)
        return sql, params

    async def data(self, db_info, export=False):
        limit = request.args.get('_size', ROWS_LIMIT) if not export else -1
        rowid = not (request.args.get('_rowid') == 'hide') and not export
        total = not (request.args.get('_total') == 'hide') and not export
        sort = request.args.get('_sort')
        sort_desc = request.args.get('_sort_desc')
        offset = request.args.get('_offset') if not export else 0

        # get filter arguments, like column__exact=xxx
        filters = []
        for key, value in request.args.items():
            if not key.startswith('_') and '__' in key:
                filters.append((key, value))

        cols = 'rowid, *' if rowid else '*'
        sql = 'SELECT {} FROM [{}]'.format(cols, db_info['table_name'])
        sql, params = self.add_filters_to_sql(sql, filters)
        if sort:
            sql += f' ORDER BY [{sort}]'
        elif sort_desc:
            sql += f' ORDER BY [{sort_desc}] DESC'
        else:
            sql += ' ORDER BY rowid'
        sql += ' LIMIT :l'
        params['l'] = limit
        if offset:
            sql += ' OFFSET :o'
            params['o'] = offset
        rows, description = await self.execute(
            sql, db_info, params=params
        )

        columns = [r[0] for r in description]

        if export:
            return columns, rows

        res = {
            'columns': columns,
            'rows': list(rows),
        }

        if total:
            sql = f"SELECT COUNT(*) FROM [{db_info['table_name']}]"
            sql, params = self.add_filters_to_sql(sql, filters)
            r, _ = await self.execute(sql, db_info, params=params)
            res['total'] = r[0][0]

        return res

    async def get(self, urlhash):
        db_info = get_db_info(urlhash)
        p = Path(db_info['db_path'])
        if not p.exists():
            raise APIError('Database has probably been removed.', status=404)

        start = time.time()
        try:
            data = await self.data(db_info)
        except (sqlite3.OperationalError, sqlite3.IntegrityError) as e:
            raise APIError('Error selecting data', status=400, payload=dict(details=str(e)))
        end = time.time()

        _shape = request.args.get('_shape', DEFAULT_SHAPE)
        if _shape == 'objects':
            # Format data as an array of objects for the client
            rows = []
            for row in data['rows']:
                rows.append(dict(zip(data['columns'], row)))
        elif _shape == 'lists':
            rows = data['rows']
        else:
            raise APIError(f"Unknown _shape: {_shape}", status=400)

        general_infos = await self.general_infos(db_info)
        columns_infos = await self.columns_infos(db_info)

        res = {
            'ok': True,
            'query_ms': (end - start) * 1000,
            'rows': rows,
            'columns': data['columns'],
            'general_infos': general_infos,
            'columns_infos': columns_infos
        }
        if data.get('total'):
            res['total'] = data['total']

        return jsonify(res)

    async def general_infos(self, db_info):
        params = {}
        sql = 'SELECT count(*) FROM sqlite_master WHERE type=\'table\' AND name=\'general_infos\''
        rows, description = await self.execute(
            sql, db_info, params=params
        )
        if rows[0][0] != 0:
            sql = 'SELECT * FROM general_infos'
            rows, description = await self.execute(
                sql, db_info, params=params
            )
            columns = [r[0] for r in description]
            res = {}
            cpt = 0
            for col in columns:
                res[col] = rows[0][cpt]
                cpt = cpt + 1

            return res
        else:
            return {}

    async def columns_infos(self, db_info):
        params = {}
        sql = 'SELECT count(*) FROM sqlite_master WHERE type=\'table\' AND name=\'columns_infos\''
        rows, description = await self.execute(
            sql, db_info, params=params
        )
        if rows[0][0] != 0:
            sql = 'SELECT * FROM columns_infos'
            rows, description = await self.execute(
                sql, db_info, params=params
            )
            columns = [r[0] for r in description]

            res = {}
            for row in rows:
                cpt = 1
                res[row[0]] = {}
                for col in columns[1:]:
                    res[row[0]][col] = row[cpt]
                    cpt = cpt + 1

            res = await self.top_and_categorical_infos(db_info, res, 'top_infos')
            res = await self.top_and_categorical_infos(db_info, res, 'categorical_infos')
            res = await self.numeric_infos(db_info, res)
            res = await self.numeric_plot_infos(db_info, res)
            return res
        else:
            return {}

    async def top_and_categorical_infos(self, db_info, res, table_name):
        params = {}
        sql = 'SELECT count(*) FROM sqlite_master WHERE type=\'table\' AND name=\'{}\''.format(table_name)
        rows, description = await self.execute(
            sql, db_info, params=params
        )
        if rows[0][0] != 0:
            sql = 'SELECT * FROM {}'.format(table_name)
            rows, description = await self.execute(
                sql, db_info, params=params
            )

            for row in rows:
                if table_name not in res[row[0]]:
                    res[row[0]][table_name] = []
                inter = {}
                inter['value'] = row[1]
                inter['count'] = row[2]
                res[row[0]][table_name].append(inter)

            return res
        else:
            for col in res:
                res[col][table_name] = {}
            return res

    async def numeric_infos(self, db_info, res):
        params = {}
        sql = 'SELECT count(*) FROM sqlite_master WHERE type=\'table\' AND name=\'numeric_infos\''
        rows, description = await self.execute(
            sql, db_info, params=params
        )
        if rows[0][0] != 0:
            sql = 'SELECT * FROM {}'.format('numeric_infos')
            rows, description = await self.execute(
                sql, db_info, params=params
            )

            for row in rows:
                if 'numeric_infos' not in res[row[0]]:
                    res[row[0]]['numeric_infos'] = {}

                res[row[0]]['numeric_infos']['mean'] = row[1]
                res[row[0]]['numeric_infos']['std'] = row[2]
                res[row[0]]['numeric_infos']['min'] = row[3]
                res[row[0]]['numeric_infos']['max'] = row[4]

            return res
        else:
            for col in res:
                res[col]['numeric_infos'] = {}
            return res

    async def numeric_plot_infos(self, db_info, res):
        params = {}
        sql = 'SELECT count(*) FROM sqlite_master WHERE type=\'table\' AND name=\'numeric_plot_infos\''
        rows, description = await self.execute(
            sql, db_info, params=params
        )
        if rows[0][0] != 0:
            sql = 'SELECT * FROM {}'.format('numeric_plot_infos')
            rows, description = await self.execute(
                sql, db_info, params=params
            )

            for row in rows:
                if 'numeric_plot_infos' not in res[row[0]]:
                    res[row[0]]['numeric_plot_infos'] = {}
                if 'counts' not in res[row[0]]['numeric_plot_infos']:
                    res[row[0]]['numeric_plot_infos']['counts'] = []
                if 'bin_edges' not in res[row[0]]['numeric_plot_infos']:
                    res[row[0]]['numeric_plot_infos']['bin_edges'] = []
                if row[2] == 'counts':
                    res[row[0]]['numeric_plot_infos']['counts'].append(row[1])
                if row[2] == 'bin_edges':
                    res[row[0]]['numeric_plot_infos']['bin_edges'].append(row[1])

            return res
        else:
            for col in res:
                res[col]['numeric_plot_infos'] = {}
            return res



import os
import tempfile

import aiohttp
import validators
import pandas as pd

from quart import request, jsonify, current_app as app
from quart.views import MethodView

from csvapi.profileview import ProfileView

from csvapi.errors import APIError
from csvapi.parser import parse
from csvapi.utils import (
    already_exists,
    get_hash,
    check_csv_detective_report_structure,
    check_profile_report_structure,
    create_connection,
    enrich_db_with_metadata
)

from csv_detective.explore_csv import routine


class ParseView(MethodView):

    @staticmethod
    async def do_parse(
        url,
        urlhash,
        encoding,
        storage,
        logger,
        sniff_limit,
        max_file_size,
        agate_types=None,
        analysis=None
    ):
        logger.debug('* do_parse %s (%s)', urlhash, url)
        tmp = tempfile.NamedTemporaryFile(delete=False)
        chunk_count = 0
        chunk_size = 1024
        try:
            async with aiohttp.ClientSession(raise_for_status=True) as session:
                async with session.get(url) as resp:
                    while True:
                        chunk = await resp.content.read(chunk_size)
                        if chunk_count * chunk_size > max_file_size:
                            tmp.close()
                            raise Exception('File too big (max size is %s bytes)' % max_file_size)
                        if not chunk:
                            break
                        tmp.write(chunk)
                        chunk_count += 1
            tmp.close()

            logger.debug('* Downloaded %s', urlhash)
            logger.debug('* Parsing %s...', urlhash)
            is_csv = parse(
                tmp.name,
                urlhash,
                storage,
                encoding=encoding,
                sniff_limit=sniff_limit,
                agate_types=agate_types
            )

            if is_csv and analysis and analysis == 'yes':
                csv_detective_report = routine(tmp.name)

                if not check_csv_detective_report_structure(csv_detective_report):
                    logger.error(
                        "csvdetective report malformed"
                    )
                    return

                profileViewInstance = ProfileView()
                profile_report = await profileViewInstance.get_minimal_profile(
                    profileViewInstance,
                    urlhash=urlhash,
                    csv_detective_report=csv_detective_report
                )

                if not check_profile_report_structure(profile_report):
                    logger.error(
                        "pandas profiling report malformed"
                    )
                    return

                enrich_db_with_metadata(
                    urlhash,
                    csv_detective_report,
                    profile_report,
                    None,
                    None
                )

            if not is_csv and analysis and analysis == 'yes':
                conn = create_connection(app.config['DB_ROOT_DIR'] + '/' + urlhash + '.db')
                general_infos = [
                    {
                        'filetype': 'excel'
                    }
                ]
                df = pd.DataFrame(general_infos)
                df.to_sql('general_infos', con=conn, if_exists='replace', index=False)

            logger.debug('* Parsed %s', urlhash)
        finally:
            logger.debug('Removing tmp file: %s', tmp.name)
            os.unlink(tmp.name)

    async def get(self):
        app.logger.debug('* Starting ParseView.get')
        url = request.args.get('url')
        encoding = request.args.get('encoding')
        if not url:
            raise APIError('Missing url query string variable.', status=400)
        if not validators.url(url):
            raise APIError('Malformed url parameter.', status=400)
        urlhash = get_hash(url)
        analysis = request.args.get('analysis')
        if not await already_exists(urlhash, analysis):
            try:
                storage = app.config['DB_ROOT_DIR']
                await self.do_parse(url=url,
                                    urlhash=urlhash,
                                    encoding=encoding,
                                    storage=storage,
                                    logger=app.logger,
                                    sniff_limit=app.config.get('CSV_SNIFF_LIMIT'),
                                    max_file_size=app.config.get('MAX_FILE_SIZE'),
                                    analysis=analysis)
            except Exception as e:
                raise APIError('Error parsing CSV: %s' % e)
        else:
            app.logger.info(f"{urlhash}.db already exists, skipping parse.")
        scheme = 'https' if app.config.get('FORCE_SSL') else request.scheme
        return jsonify({
            'ok': True,
            'endpoint': f"{scheme}://{request.host}/api/{urlhash}",
            'profile_endpoint': f"{scheme}://{request.host}/profile/{urlhash}",
        })



import os

import agate
import cchardet as chardet

from csvapi.utils import get_db_info
from csvapi.type_tester import agate_tester

from quart import current_app as app

SNIFF_LIMIT = 4096
CSV_FILETYPES = ('text/plain', 'application/csv', 'text/csv')


def detect_type(filepath):
    with os.popen(f'file {filepath} -b --mime-type') as proc:
        return proc.read().lower()


def detect_encoding(filepath):
    with open(filepath, 'rb') as f:
        return chardet.detect(f.read()).get('encoding')


def from_csv(filepath, encoding='utf-8', sniff_limit=SNIFF_LIMIT, agate_types=None):
    """Try first w/ sniffing and then w/o sniffing if it fails,
    and then again by forcing ';' delimiter w/o sniffing"""
    if not agate_types:
        kwargs = {
            'sniff_limit': sniff_limit,
            'encoding': encoding,
            'column_types': agate_tester()
        }
    else:
        kwargs = {
            'sniff_limit': sniff_limit,
            'encoding': encoding,
            'column_types': agate_types
        }
    try:
        return agate.Table.from_csv(filepath, **kwargs)
    except ValueError:
        try:
            kwargs.pop('sniff_limit')
            return agate.Table.from_csv(filepath, **kwargs)
        except ValueError:
            kwargs['delimiter'] = ';'
            return agate.Table.from_csv(filepath, **kwargs)
    except agate.CastError:
        try:
            kwargs = {
                'sniff_limit': sniff_limit,
                'encoding': encoding,
                'column_types': agate_tester()
            }
            app.logger.warning('Types from csv-detective provoke errors, use of agate type tester instead.')
            return agate.Table.from_csv(filepath, **kwargs)
        except Exception as e:
            app.logger.error('error casting %s', e)


def from_excel(filepath, xlsx=False):
    # Function exists to prevent side-effects after monckey patching with import
    import agateexcel  # noqa
    if xlsx:
        return agate.Table.from_xlsx(filepath, column_types=agate_tester())
    return agate.Table.from_xls(filepath, column_types=agate_tester())


def to_sql(table, urlhash, storage):
    db_info = get_db_info(urlhash, storage=storage)
    table.to_sql(db_info['dsn'], db_info['db_name'], overwrite=True)


def parse(filepath, urlhash, storage, encoding=None, sniff_limit=SNIFF_LIMIT, agate_types=None):
    is_csv = False
    file_type = detect_type(filepath)
    if 'application/vnd.ms-excel' in file_type:
        table = from_excel(filepath)
    elif 'application/vnd.openxml' in file_type:
        table = from_excel(filepath, xlsx=True)
    elif any([supported in file_type for supported in CSV_FILETYPES]):
        encoding = detect_encoding(filepath) if not encoding else encoding
        table = from_csv(filepath, encoding=encoding, sniff_limit=sniff_limit, agate_types=agate_types)
        is_csv = True
    else:
        raise Exception(f'Unsupported file type {file_type}')
    to_sql(table, urlhash, storage)
    return is_csv



import hashlib

from pathlib import Path

from quart import current_app as app

import sqlite3
from datetime import datetime
import pandas as pd

executor = None


def get_db_info(urlhash, storage=None):
    if app:
        # app.config not thread safe, sometimes we need to pass storage directly
        db_storage = storage or app.config['DB_ROOT_DIR']
        profile_storage = app.config['PROFILES_ROOT_DIR']

    db_path = f"{db_storage}/{urlhash}.db"
    profile_path = f"{profile_storage}/{urlhash}.html"
    return {
        'dsn': f"sqlite:///{db_path}",
        'db_name': urlhash,
        'table_name': urlhash,
        'db_path': db_path,
        'profile_path': profile_path,
    }


def get_hash(to_hash):
    return get_hash_bytes(to_hash.encode('utf-8'))


def get_hash_bytes(to_hash):
    return hashlib.md5(to_hash).hexdigest()


async def already_exists(urlhash, analysis=None):
    '''
    Check if db exist. If analysis is requested, we check if general_infos table exist.
    If not, we bypass cache and do a new download of file to analyse it with pp and csv-detective.
    '''
    cache_enabled = app.config.get('CSV_CACHE_ENABLED')
    if not cache_enabled:
        return False

    db_exist = Path(get_db_info(urlhash)['db_path']).exists()

    if not analysis or analysis != 'yes':
        return db_exist
    else:
        conn = create_connection(get_db_info(urlhash)['db_path'])
        cur = conn.cursor()
        sql = 'SELECT count(*) FROM sqlite_master WHERE type=\'table\' AND name=\'general_infos\''
        cur.execute(sql)
        rows = cur.fetchall()
        if rows[0][0] != 0:
            return True
        else:
            return False


def create_connection(db_file):
    conn = None
    conn = sqlite3.connect(db_file)
    return conn


def keys_exists(element, *keys):
    '''
    Check if *keys (nested) exists in `element` (dict).
    '''
    if not isinstance(element, dict):
        raise AttributeError('keys_exists() expects dict as first argument.')
    if len(keys) == 0:
        raise AttributeError('keys_exists() expects at least two arguments, one given.')
    _element = element
    for key in keys:
        try:
            _element = _element[key]
        except KeyError:
            return False
    return True


def check_csv_detective_report_structure(report):
    if (report is not None) and \
            (keys_exists(report, "columns")) and \
            (keys_exists(report, "encoding")) and \
            (keys_exists(report, "separator")) and \
            (keys_exists(report, "header_row_idx")):

        for item in report['columns']:
            if (not keys_exists(report, "columns", item, "python_type")) | \
                    (not keys_exists(report, "columns", item, "format")):
                return False
        return True
    else:
        return False


def check_profile_report_structure(report):
    if (report is not None) and \
            (keys_exists(report, "table", "n")) and \
            (keys_exists(report, "table", "n_var")) and \
            (keys_exists(report, "table", "n_cells_missing")) and \
            (keys_exists(report, "table", "n_vars_with_missing")) and \
            (keys_exists(report, "table", "n_vars_all_missing")) and \
            (keys_exists(report, "table", "n_cells_missing")) and \
            (keys_exists(report, "variables")):

        for item in report['variables']:
            if (not keys_exists(report, "variables", item, "n_distinct")) | \
                    (not keys_exists(report, "variables", item, "is_unique")) | \
                    (not keys_exists(report, "variables", item, "n_unique")) | \
                    (not keys_exists(report, "variables", item, "type")) | \
                    (not keys_exists(report, "variables", item, "n_missing")) | \
                    (not keys_exists(report, "variables", item, "count")):
                return False
        return True
    else:
        return False


def df_to_sql(obj, conn, name):
    df = pd.DataFrame(obj)
    if df.shape[0] > 0:
        df.to_sql(name, con=conn, if_exists='replace', index=False)


def enrich_db_with_metadata(urlhash, csv_detective_report, profile_report, dataset_id, key):
    # Save to sql
    conn = create_connection(app.config['DB_ROOT_DIR'] + '/' + urlhash + '.db')

    general_infos = [
        {
            'encoding': csv_detective_report['encoding'],
            'separator': csv_detective_report['separator'],
            'header_row_idx': csv_detective_report['header_row_idx'],
            'total_lines': profile_report['table']['n'],
            'nb_columns': profile_report['table']['n_var'],
            'nb_cells_missing': profile_report['table']['n_cells_missing'],
            'nb_vars_with_missing': profile_report['table']['n_vars_with_missing'],
            'nb_vars_all_missing': profile_report['table']['n_vars_all_missing'],
            'date_last_check': datetime.today().strftime('%Y-%m-%d'),
            'dataset_id': dataset_id,
            'resource_id': key,
            'filetype': 'csv'
        }
    ]
    df = pd.DataFrame(general_infos)
    df.to_sql('general_infos', con=conn, if_exists='replace', index=False)

    columns_infos = []
    categorical_infos = []
    top_infos = []
    numeric_infos = []
    numeric_plot_infos = []
    for col in profile_report['variables']:
        column_info = {}
        column_info['name'] = col
        column_info['format'] = csv_detective_report['columns'][col]['format']
        column_info['nb_distinct'] = profile_report['variables'][col]['n_distinct']
        column_info['is_unique'] = profile_report['variables'][col]['is_unique']
        column_info['nb_unique'] = profile_report['variables'][col]['n_unique']
        column_info['type'] = profile_report['variables'][col]['type']
        column_info['nb_missing'] = profile_report['variables'][col]['n_missing']
        column_info['count'] = profile_report['variables'][col]['count']
        columns_infos.append(column_info)

        if csv_detective_report['columns'][col]['format'] in \
                ['siren', 'siret', 'code_postal', 'code_commune_insee', 'code_departement', 'code_region', 'tel_fr']:
            column_info['type'] = 'Categorical'

        if (column_info['type'] == 'Categorical') & \
                (len(profile_report['variables'][col]['value_counts_without_nan']) < 10):
            for cat in profile_report['variables'][col]['value_counts_without_nan']:
                categorical_info = {}
                categorical_info['column'] = col
                categorical_info['value'] = cat
                categorical_info['nb'] = profile_report['variables'][col]['value_counts_without_nan'][cat]
                categorical_infos.append(categorical_info)

        if column_info['type'] == 'Numeric':
            numeric_info = {}
            numeric_info['column'] = col
            numeric_info['mean'] = profile_report['variables'][col]['mean']
            numeric_info['std'] = profile_report['variables'][col]['std']
            numeric_info['min'] = profile_report['variables'][col]['min']
            numeric_info['max'] = profile_report['variables'][col]['max']
            numeric_infos.append(numeric_info)
            for i in range(len(profile_report['variables'][col]['histogram']['bin_edges'])):
                numeric_plot_info = {}
                numeric_plot_info['column'] = col
                numeric_plot_info['value'] = profile_report['variables'][col]['histogram']['bin_edges'][i]
                numeric_plot_info['type'] = 'bin_edges'
                numeric_plot_infos.append(numeric_plot_info)

            for i in range(len(profile_report['variables'][col]['histogram']['counts'])):
                numeric_plot_info = {}
                numeric_plot_info['column'] = col
                numeric_plot_info['value'] = profile_report['variables'][col]['histogram']['counts'][i]
                numeric_plot_info['type'] = 'counts'
                numeric_plot_infos.append(numeric_plot_info)

        cpt = 0
        for top in profile_report['variables'][col]['value_counts_without_nan']:
            if (cpt < 10):
                top_info = {}
                top_info['column'] = col
                top_info['value'] = top
                top_info['nb'] = profile_report['variables'][col]['value_counts_without_nan'][top]
                top_infos.append(top_info)
                cpt = cpt + 1

    df_to_sql(columns_infos, conn, 'columns_infos')
    df_to_sql(categorical_infos, conn, 'categorical_infos')
    df_to_sql(top_infos, conn, 'top_infos')
    df_to_sql(numeric_infos, conn, 'numeric_infos')
    df_to_sql(numeric_plot_infos, conn, 'numeric_plot_infos')

    conn.commit()



class APIError(Exception):
    status = 500

    def __init__(self, message, status=None, payload=None):
        super().__init__(message)
        self.message = message
        if status is not None:
            self.status = status
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['error'] = self.message
        rv['ok'] = False
        return rv



import os
import traceback

from quart import Quart, jsonify
from quart_cors import cors
from werkzeug.exceptions import NotFound

from csvapi.errors import APIError
from csvapi.tableview import TableView
from csvapi.exportview import ExportView
from csvapi.uploadview import UploadView
from csvapi.parseview import ParseView
from csvapi.profileview import ProfileView
from csvapi.security import filter_referrers

app = Quart(__name__)
app = cors(app, allow_origin='*')

app.add_url_rule('/api/<urlhash>', view_func=TableView.as_view('table'))
app.add_url_rule('/profile/<urlhash>', view_func=ProfileView.as_view('profile'))
app.add_url_rule('/api/<urlhash>/export', view_func=ExportView.as_view('export'))
app.add_url_rule('/apify', view_func=ParseView.as_view('parse'))
app.add_url_rule('/upload', view_func=UploadView.as_view('upload'))
app.before_request(filter_referrers)


conffile = os.environ.get('CSVAPI_CONFIG_FILE') or '../config.py'
app.config.from_pyfile(conffile)


def handle_and_print_error(error):
    sentry_id = None
    if app.config.get('SENTRY_DSN'):
        import sentry_sdk
        with sentry_sdk.push_scope() as scope:
            sentry_sdk.init(
                app.config['SENTRY_DSN'],
                traces_sample_rate=1.0
            )
            scope.set_extra('debug', False)
            from sentry_sdk import capture_exception
            sentry_id = capture_exception(error)
    traceback.print_exc()
    return sentry_id


@app.errorhandler(NotFound)
def handle_not_found(error):
    response = jsonify({
        'ok': False,
        'error': 'Not found',
    })
    response.status_code = 404
    return response


@app.errorhandler(APIError)
def handle_api_error(error):
    error_id = handle_and_print_error(error)
    data = error.to_dict()
    app.logger.error(f"{data.get('error')}: {data.get('details', '')}")
    data['error_id'] = error_id
    response = jsonify(data)
    response.status_code = error.status
    return response


@app.errorhandler(Exception)
def handle_exceptions(error):
    """Serialize all errors to API"""
    error_id = handle_and_print_error(error)
    response = jsonify(error=str(error), error_id=error_id, ok=False)
    return response, 500



import click
import ssl
from click_default_group import DefaultGroup

from csvapi.webservice import app

RESPONSE_TIMEOUT = 5 * 60  # in seconds


@click.group(cls=DefaultGroup, default='serve', default_if_no_args=True)
@click.version_option()
def cli():
    """
    csvapi!
    """


@click.option('--dbs', default='./dbs',
              type=click.Path(exists=True, file_okay=False),
              help='Where to store sqlite DBs')
@click.option('-h', '--host', default='127.0.0.1',
              help='host for server, defaults to 127.0.0.1')
@click.option('-p', '--port', default=8001,
              help='port for server, defaults to 8001')
@click.option('--debug', is_flag=True,
              help='Enable debug mode - useful for development')
@click.option('--reload', is_flag=True,
              help='Automatically reload if code change detected')
@click.option('--cache/--no-cache', default=True,
              help='Do not parse CSV again if DB already exists')
@click.option('-w', '--max-workers', default=3,
              help='Max number of ThreadPoolExecutor workers')
@click.option('--ssl-cert', default=None,
              help='Path to SSL certificate')
@click.option('--ssl-key', default=None,
              help='Path to SSL key')
@cli.command()
def serve(dbs, host, port, debug, reload, cache, max_workers, ssl_cert, ssl_key):
    ssl_context = None
    if ssl_cert and ssl_key:
        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_context.load_cert_chain(certfile=ssl_cert, keyfile=ssl_key)
    app.config.update({
        'DB_ROOT_DIR': dbs,
        'CSV_CACHE_ENABLED': cache,
        'MAX_WORKERS': max_workers,
        'DEBUG': debug,
        # TODO this probably does not exist in Quart
        'RESPONSE_TIMEOUT': RESPONSE_TIMEOUT,
    })
    app.run(host=host, port=port, debug=debug, use_reloader=reload, ssl=ssl_context)



import re

from agate.data_types.base import DataType
from agate.data_types.boolean import Boolean
from agate.data_types.date import Date
from agate.data_types.date_time import DateTime
from agate.data_types.number import Number
from agate.data_types.text import Text
from agate.data_types.time_delta import TimeDelta
from agate.exceptions import CastError
from agate.type_tester import TypeTester

from agatesql import table as agatesqltable

from sqlalchemy.types import VARCHAR

from stdnum.fr.siren import is_valid as is_valid_siren
from stdnum.fr.siret import is_valid as is_valid_siret


class Time(DataType):
    # Detect an hour minute string.
    # Examples: 12:20, 9:50, 23:30
    def __init__(self, **kwargs):
        super(Time, self).__init__(**kwargs)

    def cast(self, d):
        if d is None:
            return d
        if re.match(r"^(?:[01]\d|2[0-3]|\d):[0-5]\d$", str(d)):
            return Text().cast(d)
        raise CastError('Can not parse value "%s" as time.' % d)


class SirenSiret(DataType):
    # Detect a SIREN or SIRET number
    def __init__(self):
        super(SirenSiret, self).__init__()

    def cast(self, d):
        if d is None:
            return d
        if is_valid_siret(d) or is_valid_siren(d):
            return Text().cast(d)
        raise CastError('Can not parse value "%s" as a SIREN or SIRET.' % d)


# agatesql needs to know the SQL equivalent of a type.
# Tell agatesql how our custom types should be converted in SQL.
#
# Reference:
# https://github.com/wireservice/agate-sql/blob/7466073d81289323851c21817ea33170e36ce2a5/agatesql/table.py#L21-L28
agatesqltable.SQL_TYPE_MAP[Time] = VARCHAR
agatesqltable.SQL_TYPE_MAP[SirenSiret] = VARCHAR


def agate_tester():
    # Override the original list of type checkers present in agate
    # to detect types.
    #
    # Original list here:
    # https://github.com/wireservice/agate/blob/e3078dca8b3566e8408e65981f79918c2f36f9fe/agate/type_tester.py#L64-L71
    return TypeTester(
        types=[
            Boolean(),
            SirenSiret(),
            Number(),
            Time(),
            TimeDelta(),
            Date(),
            DateTime(),
            Text(),
        ]
    )


def convert_types(column_types):
    convert_dict = {
        'string': Text(),
        'int': Number(),
        'float': Number(),
        'number': Number(),
        'date': Date(),
        # 'bool': Boolean(),
        'bool': Text(),
        'geopoint': Text(),
        'geojson': Text(),
        'any': Text(),
        'year': Text(),
        'datetime': DateTime(),
        'time': Time()
    }
    agate_types = []
    for ct in column_types:
        agate_types.append(convert_dict[ct])

    return agate_types


def convert_python_types(columns):
    convert_dict = {
        'string': str,
        'int': int,
        'float': float,
        'number': float,
        'date': str,
        # 'bool': bool,
        'bool': str,
        'geopoint': str,
        'geojson': str,
        'any': str,
        'year': str,
        'datetime': str,
        'time': str
    }
    python_types = {}
    for col in columns:
        python_types[col] = convert_dict[columns[col]['python_type']]

    return python_types



import os
from tempfile import NamedTemporaryFile

from quart import request, current_app as app, jsonify
from quart.views import MethodView

from csvapi.errors import APIError
from csvapi.utils import get_hash_bytes, already_exists
from csvapi.parser import parse


class UploadView(MethodView):

    async def post(self):
        files = await request.files
        _file = files.get('file') or files.get('filepond')
        if not _file:
            raise APIError('Missing file.', status=400)
        content_hash = get_hash_bytes(_file.read())
        _file.seek(0)
        if not already_exists(content_hash):
            storage = app.config['DB_ROOT_DIR']
            sniff_limit = app.config.get('CSV_SNIFF_LIMIT')
            try:
                _tmpfile = NamedTemporaryFile(delete=False)
                _file.save(_tmpfile)
                _tmpfile.close()
                parse(_tmpfile.name, content_hash, storage, sniff_limit=sniff_limit)
            finally:
                os.unlink(_tmpfile.name)

        scheme = 'https' if app.config.get('FORCE_SSL') else request.scheme
        return jsonify({
            'ok': True,
            'endpoint': f"{scheme}://{request.host}/api/{content_hash}"
        })



import csv
import sqlite3

from io import StringIO
from pathlib import Path

from quart import make_response

from csvapi.errors import APIError
from csvapi.tableview import TableView
from csvapi.utils import get_db_info


class ExportView(TableView):

    async def get(self, urlhash):
        "This will inherit sorting and filtering from TableView"
        db_info = get_db_info(urlhash)
        p = Path(db_info['db_path'])
        if not p.exists():
            raise APIError('Database has probably been removed.', status=404)

        try:
            columns, rows_iter = await self.data(db_info, export=True)
        except (sqlite3.OperationalError, sqlite3.IntegrityError) as e:
            raise APIError('Error selecting data', status=400, payload=dict(details=str(e)))

        def make_line(line_data):
            line = StringIO()
            writer = csv.writer(line)
            writer.writerow(line_data)
            line.seek(0)
            return line.read().encode()

        async def _make_response():
            yield make_line(columns)
            for line in rows_iter:
                yield make_line(line)

        response = await make_response(_make_response())
        response.mimetype = 'text/csv'
        response.headers['Content-Disposition'] = f'attachment; filename={urlhash}.csv'
        return response



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['csvapi']

package_data = \
{'': ['*']}

install_requires = \
['Quart>=0.18.0,<0.19.0',
 'agate-excel>=0.2.5,<0.3.0',
 'agate-sql>=0.5.8,<0.6.0',
 'agate>=1.6.3,<1.7.0',
 'aiohttp>=3.8.1,<3.9.0',
 'aiosqlite>=0.17.0,<0.18.0',
 'boto3>=1.24.66,<1.25.0',
 'cchardet>=2.1.7,<2.2.0',
 'click>=8.1.3,<8.2.0',
 'click_default_group>=1.2.2,<1.3.0',
 'csv-detective>=0.4.5,<0.5.0',
 'pandas-profiling>=3.2.0,<3.3.0',
 'pandas>=1.4.4,<1.5.0',
 'python-stdnum>=1.17,<1.18',
 'quart-cors>=0.5.0,<0.6.0',
 'requests>=2.28.1,<2.29.0',
 'sentry-sdk>=1.9.8,<1.10.0',
 'validators>=0.20.0,<0.21.0']

entry_points = \
{'console_scripts': ['csvapi = csvapi.cli:cli']}

setup_kwargs = {
    'name': 'csvapi',
    'version': '2.0.1.dev716',
    'description': 'An instant JSON API for your CSV',
    'long_description': None,
    'author': 'Opendatateam',
    'author_email': 'opendatateam@data.gouv.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4',
}


setup(**setup_kwargs)



from urllib.parse import urlparse

from quart import current_app as app, request, jsonify


def filter_referrers():
    filters = app.config.get('REFERRERS_FILTER')
    if not filters:
        return None
    referrer = request.referrer
    if referrer:
        parsed = urlparse(referrer)
        for filter in filters:
            if parsed.hostname.endswith(filter):
                return None
    return jsonify({
            'ok': False,
            'error': 'Unauthorized',
    }), 403



from pathlib import Path

import pandas as pd
import sqlite3

from quart import send_from_directory
from quart.views import MethodView
from pandas_profiling import ProfileReport

from csvapi.errors import APIError
from csvapi.utils import get_db_info

from csvapi.type_tester import convert_python_types

from quart import current_app as app

import json


class ProfileView(MethodView):

    def get_dataframe(self, db_info, dtype=None):
        dsn = 'file:{}?immutable=1'.format(db_info['db_path'])
        conn = sqlite3.connect(dsn, uri=True)
        sql = 'SELECT * FROM [{}]'.format(db_info['table_name'])
        try:
            df = pd.read_sql_query(sql, con=conn, dtype=dtype)
        # TODO: check if correct exception type
        except ValueError:
            df = pd.read_sql_query(sql, con=conn)
            app.logger.debug('problem with python types')
        return df

    def make_profile(self, db_info):
        df = self.get_dataframe(db_info)

        if app.config['PANDAS_PROFILING_CONFIG_MIN']:
            profile = ProfileReport(df, config_file="profiling-minimal.yml")
        else:
            profile = ProfileReport(df)
        profile.to_file(db_info['profile_path'])
        return Path(db_info['profile_path'])

    async def get(self, urlhash):
        db_info = get_db_info(urlhash)
        p = Path(db_info['db_path'])
        if not p.exists():
            raise APIError('Database has probably been removed or does not exist yet.', status=404)

        path = Path(db_info['profile_path'])

        if not path.exists():
            try:
                path = self.make_profile(db_info)
            except (sqlite3.OperationalError, sqlite3.IntegrityError) as e:
                raise APIError('Error selecting data', status=400, payload=dict(details=str(e)))

        return await send_from_directory(path.parent, path.name)

    async def get_minimal_profile(self, url: str, urlhash: str, csv_detective_report: dict) -> None:
        db_info = get_db_info(urlhash)
        p = Path(db_info['db_path'])
        if not p.exists():
            raise APIError('Database has probably been removed or does not exist yet.', status=404)

        path = Path(db_info['profile_path'])

        if not path.exists():
            try:

                python_types = convert_python_types(csv_detective_report['columns'])
                df = self.get_dataframe(db_info, dtype=python_types)
                profile = ProfileReport(
                    df, minimal=True,
                    vars=dict(num={"low_categorical_threshold": 0}),
                    plot=dict(histogram={"bins": 10})
                )
                profile_report = json.loads(profile.to_json())
                return profile_report
            except (sqlite3.OperationalError, sqlite3.IntegrityError) as e:
                raise APIError('Error selecting data', status=400, payload=dict(details=str(e)))

        return await send_from_directory(path.parent, path.name)
